"""Provide fixtures for the entire parent directory."""

from tests.fixtures.alphatracker import *  # noqa: F403
from tests.fixtures.camera import *  # noqa: F403
from tests.fixtures.coco import *  # noqa: F403
from tests.fixtures.cvat import *  # noqa: F403
from tests.fixtures.dlc import *  # noqa: F403
from tests.fixtures.jabs import *  # noqa: F403
from tests.fixtures.labels import *  # noqa: F403
from tests.fixtures.labelstudio import *  # noqa: F403
from tests.fixtures.leap import *  # noqa: F403
from tests.fixtures.slp import *  # noqa: F403
from tests.fixtures.ultralytics import *  # noqa: F403
from tests.fixtures.videos import *  # noqa: F403
